package net.sf.javaprinciples.data.transformer;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;

import au.com.sparxsystems.Association;
import au.com.sparxsystems.AssociationProperties;
import au.com.sparxsystems.Attribute;
import au.com.sparxsystems.Element;
import net.sf.jcc.model.parser.UnexpectedException;
import net.sf.jcc.model.parser.uml2.ElementStore;
import net.sf.jcc.model.parser.uml2.ModelElement;

/**
 * Abstract class to implement common functionality in creating transformers or mappers from model elements.
 * @author Kay Chevalier
 */
public abstract class ModelElementFactory<T,U>
{
    public static final String IMPLEMENTATION_TAG = "implementation";
    public static final String RIGHT_STEREOTYPE_PROPERTY = "right";
    public static final String LEFT_STEREOTYPE_PROPERTY = "left";
    public static final String SOURCE_DESTINATION_DIRECTION_PROPERTY = "Source -> Destination";
    public static final String POST_PROCESSING_STEREOTYPE_PROPERTY = "postProcessor";
    public static final String PRE_PROCESSING_STEREOTYPE_PROPERTY = "preProcessor";
    public static final String DESCENDANT_STEREOTYPE_PROPERTY = "..ll..ll";
    protected ObjectTypeMapper objectTypeMapper;
    protected Logger log = LoggerFactory.getLogger(ModelElementFactory.class);
    private MapperBeanFactory<T,U> mapperBeanFactory;

    protected abstract ModelElementMapper<T,U>  createModelElementMapper(String mapperGuid, ElementStore store,
                                                             TransformerFactory.DIRECTION_TYPE direction);


    protected String getModelElementStereotype(ModelElement destinationMapper)
    {
        if (destinationMapper == null || destinationMapper.getExtension() == null)
            return null;

        Object extension = destinationMapper.getExtension();
        if (extension instanceof Element)
            return ((Element)extension).getProperties().getStereotype();
        else if (extension instanceof Attribute)
            return ((Attribute)extension).getStereotype().getStereotype();

        return null;
    }

    protected List<ModelElement> retrieveSourceFromAssociations(ModelElement transformerModelElement,
                                                          TransformerFactory.DIRECTION_TYPE direction, ElementStore store)
    {
        return retrieveModelElementFromAssociations(true, transformerModelElement, direction, store);
    }

    protected List<ModelElement> retrieveDestinationFromAssociations(ModelElement transformerModelElement,
                                                               TransformerFactory.DIRECTION_TYPE direction, ElementStore store)
    {
        return retrieveModelElementFromAssociations(false, transformerModelElement, direction, store);
    }

    /**
     * Retrieves the {@link ModelElement} from the associations of a given transformerModelElement.
     * Returns a left or right stereotyped association depending on the {@link net.sf.javaprinciples.data.transformer.TransformerFactory.DIRECTION_TYPE}
     * and the isSourceAssociation provided.
     *
     * @param isSourceAssociations - indicator specifying if it is the source or destination association we are retrieving.
     * @param transformerModelElement - the model element to retrieve the associations from
     * @param direction - the direction of the transformation
     * @param store - the store to retrieve the associations from
     * @return  the left or right associated ModelElement
     */
    protected List<ModelElement> retrieveModelElementFromAssociations(boolean isSourceAssociations, ModelElement transformerModelElement,
                                                                TransformerFactory.DIRECTION_TYPE direction, ElementStore store)
    {
        List<Association> associations = transformerModelElement.getAssociations();

        List<ModelElement> elements = new ArrayList<ModelElement>();

        if (associations == null)
            return elements;

        for (Association association : associations)
        {
            String stereotype = association.getProperties() == null ? null : association.getProperties().getStereotype();
            if (TransformerFactory.DIRECTION_TYPE.leftToRight.equals(direction))
            {
                if (isSourceAssociations && LEFT_STEREOTYPE_PROPERTY.equals(stereotype))
                {
                    elements.add(retrieveAttributeFromAssociation(association, store));
                }
                else if (!isSourceAssociations && RIGHT_STEREOTYPE_PROPERTY.equals(stereotype))
                {
                    elements.add(retrieveAttributeFromAssociation(association, store));
                }

            }
            else if (TransformerFactory.DIRECTION_TYPE.rightToLeft.equals(direction))
            {
                if (isSourceAssociations && RIGHT_STEREOTYPE_PROPERTY.equals(stereotype))
                {
                    elements.add(retrieveAttributeFromAssociation(association, store));
                }
                else if (!isSourceAssociations && LEFT_STEREOTYPE_PROPERTY.equals(stereotype))
                {
                    elements.add(retrieveAttributeFromAssociation(association, store));
                }
            }
        }

        return elements;
    }

    protected ModelElement retrieveAttributeFromAssociation(Association association, ElementStore store)
    {
        if (association.getTargetAttributeId() != null)
        {
            return store.get(association.getTargetAttributeId());
        }
        else
        {
            return store.get(association.getTargetClassId());
        }
    }

    protected List<Mapper> retrievePreProcessorFromAssociations(ModelElement modelElement,
                                                                 ElementStore store,
                                                                 TransformerFactory.DIRECTION_TYPE direction)
    {
        List<?> list = retrieveFromAssociations(modelElement, store, direction, PRE_PROCESSING_STEREOTYPE_PROPERTY);
        return (List<Mapper>)list;
    }

    protected List<Mapper> retrievePostProcessorFromAssociations(ModelElement modelElement,
                                                                       ElementStore store,
                                                                       TransformerFactory.DIRECTION_TYPE direction)
    {
        List<?> list = retrieveFromAssociations(modelElement, store, direction, POST_PROCESSING_STEREOTYPE_PROPERTY);
        return (List<Mapper>)list;
    }

    protected List<Mapper> retrieveMappersFromAssociations(ModelElement transformerModelElement,
                                                      ElementStore store,
                                                      TransformerFactory.DIRECTION_TYPE direction)
    {
        List<?> list = retrieveFromAssociations(transformerModelElement, store, direction, DESCENDANT_STEREOTYPE_PROPERTY);
        return (List<Mapper>)list;
    }

    private List<?> retrieveFromAssociations(ModelElement transformerModelElement,
                                                                       ElementStore store,
                                                                       TransformerFactory.DIRECTION_TYPE direction,
                                                                       String stereotype)
    {
        List<Association> associations = transformerModelElement.getAssociations();
        if (associations == null)
        {
            throw new UnexpectedException(String.format("Unable to perform transformation for transformer %s . " +
                    "No associated mappers found.", transformerModelElement.getGuid()));
        }


        List<ModelElementMapper> modelElementMappers = new ArrayList<ModelElementMapper>();
        for (Association association : associations)
        {
            AssociationProperties associationProperties = association.getProperties();
            if (associationProperties == null)
            {
                continue;
            }
            String associationStereotype = associationProperties.getStereotype();
            if (associationStereotype == null)
            {
                associationStereotype = DESCENDANT_STEREOTYPE_PROPERTY;
            }
            String mapperDirection = associationProperties.getDirection();
            if (mapperDirection == null)
            {
                throw new UnexpectedException(String.format("Unable to perform transformation for transformer %s . " +
                        "No direction found.", association.getIdref()));
            }
            if (SOURCE_DESTINATION_DIRECTION_PROPERTY.equals(mapperDirection) && !association.isIsNavigable()
                    && associationStereotype.equals(stereotype))
            {
                String destinationMapperId = association.getTargetClassId();
                ModelElementMapper modelElementMapper = createModelElementMapper(destinationMapperId, store, direction);
                modelElementMapper.setStore(store);
                modelElementMappers.add(modelElementMapper);
            }
        }
        return modelElementMappers;
    }

    protected Mapper<T,U> getRegisteredMapper(String mapperName)
    {
        try
        {
            return mapperBeanFactory.createMapper(mapperName);
        }
        catch (NoSuchBeanDefinitionException e)
        {
            throw e;
        }
    }


    protected ModelElementMapper<T,U> getModelElementMapper(String mapperName)
    {
        Object mapper = getRegisteredMapper(mapperName);

        if (mapper instanceof ModelElementMapper)
            return (ModelElementMapper<T,U>)mapper;
        else
            throw new UnexpectedException(String.format("Expected a ModelElementMapper type for mapperName: %s", mapperName));
    }

    protected String findAssociationStereotype(Association association)
    {
        AssociationProperties properties = association.getProperties();
        return properties != null ? properties.getStereotype() : null;
    }

    public void setMapperBeanFactory(MapperBeanFactory<T, U> mapperBeanFactory)
    {
        this.mapperBeanFactory = mapperBeanFactory;
    }

    public void setObjectTypeMapper(ObjectTypeMapper objectTypeMapper)
    {
        this.objectTypeMapper = objectTypeMapper;
    }
}
